// cmdline overridable defines
#ifndef CHUNK_SIZE
#define CHUNK_SIZE 4096
#endif
#ifndef CANVAS_SIZE
#define CANVAS_SIZE 1024 * 1024
#endif
#ifndef PALETTE_SIZE
#define PALETTE_SIZE 256
#endif

#define PARAM_SIZE 8

struct ParserState {
int index;
int width;
int height;
int color;
int cursor;
int y_offset;
int offset;
int not_aborted;
int length;
int palette_length;
int jump_offsets[6];
int params[PARAM_SIZE];
int palette[PALETTE_SIZE];
char chunk[CHUNK_SIZE];
int canvas[CANVAS_SIZE];
int repeat;
};
static struct ParserState ps;

// exported functions
extern "C" {
void* get_chunk_address()   { return &ps.chunk[0];   }
void* get_canvas_address()  { return &ps.canvas[0];  }
void* get_palette_address() { return &ps.palette[0]; }
int get_chunk_limit()       { return CHUNK_SIZE;   }
int get_canvas_limit()      { return CANVAS_SIZE;  }
int get_palette_limit()     { return PALETTE_SIZE; }

void init(int width, int height, int fill_color, int palette_length);
void decode(int length);
}

// Note: This function is the optimized version of range checking (n >= lo && n <= hi).
//       Rationale: A negative integer cast into an unsigned integer will always be greater
//       than the upper bound of the byte range (127 or 255) we can use in our sixel decoder.
//       This reduces the number of comparisons. Compilers can and usually do this, at least GCC and CLANG
//       does (@ -O2/O3 level). But let us enforce it.

inline bool check_range(int n, int lo, int hi)
{
	// hi MUST be >= lo (assert here)
	return unsigned(n - lo) < (hi - lo + 1); // faster than checking the upper and lower bound.
}

inline int peek()
{
	return ps.index < ps.length ? ps.chunk[ps.index] & 0x7F : -1;
}

inline int get()
{
	return ps.index < ps.length ? ps.chunk[ps.index++] & 0x7F : -1;
}

inline void resetparams()
{
	ps.params[0] = 0;
	ps.params[1] = 0;
	ps.params[2] = 0;
	ps.params[3] = 0;
	ps.params[4] = 0;
	ps.params[5] = 0;
	ps.params[6] = 0;
	ps.params[7] = 0;
}
// Normalize %-based SIXEL RGB 0..100 to channel byte values 0..255.
// Note: does some rounding in integer arithmetics.
inline int NormalizeRGB(int r, int g, int b) {
return 0xFF000000 | ((b * 255 + 99) / 100) << 16 | ((g * 255 + 99) / 100) << 8 | ((r * 255 + 99) / 100);
}

// hue to channel value helper.
inline float h2c(float t1, float t2, float c) {
if (c < 0) c += 1;
else if (c > 1) c -= 1;
return c * 6 < 1
	? t2 + (t1 - t2) * 6 * c
	: c * 2 < 1
	? t1
	: c * 3 < 2
		? t2 + (t1 - t2) * (4 - c * 6)
		: t2;
}

// Normalize SIXEL HLS to RGBA8888.
// Incoming values are integer in: H - 0..360 (hue turned by 240°), L - 0..100, S - 0..100.
inline int NormalizeHLS(int hi, int li, int si) {
if (!si) {
	return 0xFF000000 | ((li * 255 + 99) / 100) << 16 | ((li * 255 + 99) / 100) << 8 | ((li * 255 + 99) / 100);
}
float h = ((float) (hi + 240 % 360)) / 360;
float l = ((float) li) / 100;
float s = ((float) si) / 100;
float t1 = l < 0.5 ? l * (1 + s) : l + s - l * s;
float t2 = l * 2 - t1;
unsigned char r = 255 * h2c(t1, t2, h + (1.0f / 3));
unsigned char g = 255 * h2c(t1, t2, h);
unsigned char b = 255 * h2c(t1, t2, h - (1.0f / 3));
return 0xFF000000 | b << 16 | g << 8 | r;
}

inline int readparams()
{
	resetparams();
	int c = 0, n = 0, i = 0;
	for(;;) {
		while((c = peek()) >= '0' && c <= '9')
			n = n * 10 + (get() - '0');
		ps.params[i++ & (PARAM_SIZE - 1)] = n;
		if(c != ';')
			break;
		n = 0;
		++ps.index;
	}
	return i;
}

inline void Return()
{
	ps.cursor = 0;
}

inline void LineFeed()
{
	ps.y_offset += 6;
	ps.offset = ps.y_offset * ps.width;
	ps.cursor = 0;
}

inline void GetRepeatCount()
{
	(void) readparams();
	ps.repeat = ps.params[0] >= 1 ? ps.params[0] : 1;
}

inline void GetRasterInfo()
{
	(void) readparams(); // Not used...
}

inline void SetPalette()
{
	int n = readparams();
	if(n == 5) {
		switch(ps.params[1]) {
		case 2: // RGB
			ps.color = ps.palette[ps.params[0] % ps.palette_length] = NormalizeRGB(ps.params[2], ps.params[3], ps.params[4]);
			break;
		case 1: // HLS
			ps.color = ps.palette[ps.params[0] % ps.palette_length] = NormalizeHLS(ps.params[2], ps.params[3], ps.params[4]);
			break;
		//case 0:
		//	ps.color = ps.palette[ps.params[0] % ps.palette_length];
		//	break;
		default:
			break;
		}
	}
	else
	if(n == 1) {
		ps.color = ps.palette[ps.params[0] % ps.palette_length];
	}

}

inline void PaintSixel(int c)
{
	if(ps.cursor >= ps.width || ps.y_offset >= ps.height)
		return;
	
	if(ps.repeat == 1) {
		int p = ps.cursor + ps.offset;
		for(int n = 0; n < 6; ++n) {
			if(c & (1 << n))
				ps.canvas[p + ps.jump_offsets[n]] = ps.color;
		}
		ps.cursor++;
	}
	else {
		int n = ps.repeat;
		if(ps.cursor + n >= ps.width) {
			n = ps.width - ps.cursor;
		}
		int p = ps.cursor + ps.offset;
		for (int i = 0; i < 6; ++i) {
			if (c & (1 << i)) {
				int pp = p + ps.jump_offsets[i];
				for (int r = 0; r < n; ++r) {
					ps.canvas[pp + r] = ps.color;
				}
			}
		}
		ps.cursor += ps.repeat;
	}
	ps.repeat = 1;
}

void init(int width, int height, int fill_color, int palette_length) {
ps.not_aborted = 1;
ps.color = 0;
ps.cursor = 0;
ps.repeat = 1;
ps.y_offset = 0;
ps.offset = 0;
ps.index  = 0;
ps.length = CHUNK_SIZE;
ps.palette_length = (0 < palette_length && palette_length < PALETTE_SIZE) ? palette_length : PALETTE_SIZE;
resetparams();

// basic overflow check
// Note: The height adjustment here is needed to avoid a potential overflow in the last sixel line.
// By adjusting here to multiples of 6 we reduce the usable canvas a bit, but can avoid nasty special
// case branching in put.
if (width == 0 || width > 0xFFFF || height == 0 || height > 0xFFFF || width * (height + 5) / 6 > CANVAS_SIZE) {
	ps.not_aborted = 0;
	ps.width  = 0;
	ps.height = 0;
	return;
}

ps.width = width;
ps.height = height;

// clear canvas with fill_color
int p = 0;
for (int y = 0; y < ps.height; ++y) {
	for (int x = 0; x < ps.width; ++x) {
	ps.canvas[p + x] = fill_color;
	}
	p = p + ps.width;
}

// calc sixel pixel line jump offsets
for (int i = 0, v = 0; i < 6; ++i, v += ps.width) {
	ps.jump_offsets[i] = v;
}
}

void decode(int length)
{
	if(ps.not_aborted && ps.y_offset < ps.height && length > 0) {
		ps.length = length < CHUNK_SIZE ? length : CHUNK_SIZE;
		ps.index = 0;
		for(;;) {
			int c = get();
			switch(c) {
			case 0x21:
				GetRepeatCount();
				break;
			case 0x22:
				GetRasterInfo();
				break;
			case 0x23:
				SetPalette();
				break;
			case 0x24:
				Return();
				break;
			case 0x2D:
				LineFeed();
				break;
			case -1: // eof or err
				return;
			default:
				if(c >= 0x3F && c <= 0x7E)
					PaintSixel(c - 0x3F);
				break;
			}
		}
	}
}

